package com.gremwell.jnetbridge;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import org.jnetpcap.Pcap;
import org.jnetpcap.PcapStat;
import org.jnetpcap.nio.JBuffer;
import org.jnetpcap.nio.JMemory;
import org.jnetpcap.packet.PcapPacket;
import org.jnetpcap.protocol.JProtocol;

/**
 * This class implements a port receiving and sending network packets
 * with libpcap, through jnetpcap library.
 *
 * Upon instantiation PCAP handle gets initialized, but the packet pumping thread
 * does not get started until start().
 *
 * The clients of this class can invoke start()/stop() several times.
 * 
 * The PCAP handle gets closed upon invocation of close() method. After that it
 * does not get reopened by start().
 *
 * @author Alexandre Bezroutchko
 * @author Gremwell bvba
 */
public class PcapPort extends Port {

    private Thread pumpThread = null;
    private Pcap pcap;
    private final Object pcapLock = new Object();
    private final PcapPacket pcapPacket;
    private BlockingQueue<JBuffer> egressPackets = new LinkedBlockingQueue<JBuffer>();
    //
    private final static int snaplen = 64 * 1024;
    private final static int flags = Pcap.MODE_PROMISCUOUS;
    private final static int timeoutMs = 1;
    private final static int bufsizeBytes = 16 * 1024 * 1024;

    private static Pcap openPcap(String ifaceName) throws PcapException {
        Pcap pcap;
        StringBuilder errbuf = new StringBuilder();

        if (!Pcap.isPcap100Loaded()) {
            throw new PcapException("jnetpcap1.4 + libpcap1.0.0+ are required");
        }

        pcap = Pcap.create(ifaceName, errbuf);
        if (pcap == null) {
            throw new PcapException("failed to create pcap interface " + ifaceName + ": " + errbuf.toString());
        }

        // standard properties
        pcap.setSnaplen(snaplen);
        pcap.setPromisc(flags);
        pcap.setTimeout(timeoutMs);

        // specific to libpcap 1.0.0
        pcap.setDirection(Pcap.Direction.IN);
        pcap.setBufferSize(bufsizeBytes);

        pcap.activate();

        return pcap;
    }

    /**
     * Prepare the interface to receive and send packets. May throw an exception if an
     * error occurs.
     *
     * @param name Name of the interface to open
     * @throws PcapException
     */
    public PcapPort(String name) throws PcapException {
        super(name);

        pcap = openPcap(name);
        pcapPacket = new PcapPacket(JMemory.Type.POINTER);
    }

    @Override
    public void send(JBuffer packet) {
        egressPackets.add(packet);
    }

    /**
     * Start packet pumping thread. The thread will run until stop() or
     * close().
     *
     */
    public synchronized void start() {
        if (pumpThread == null) {
            pumpThread = new Thread(new PcapPortPump(this));
            pumpThread.start();
        }
    }

    /**
     * Stop packet pumping thread. Invoked by close().
     *
     */
    public synchronized void stop() {
        if (pumpThread != null) {
            pumpThread.interrupt();
            pumpThread = null;
        }
    }

    public void close() {
        stop();

        synchronized (pcapLock) {
            if (pcap != null) {
                pcap.close();
                pcap = null;
            }
        }
    }

    /**
     *
     * @see Port#getStat()
     */
    @Override
    public String getStat() {
        if (pcap == null) {
            return "closed";
        }

        PcapStat ps = new PcapStat();
        pcap.stats(ps);

        return super.getStat()
                + ", egQueue=" + egressPackets.size()
                + ", pcapDrop=" + ps.getDrop()
                + ", pcapIfDrop=" + ps.getIfDrop()
                + ", pcapRecv=" + ps.getRecv();
    }

    private class PcapPortPump implements Runnable {

        public PcapPortPump(PcapPort port) {
        }

        public void run() {
            while (!Thread.interrupted()) {
                // check if jnetpcap has a packet for us
                int res;
                synchronized (pcapLock) {
                    if (pcap == null) {
                        return;
                    }
                    res = pcap.nextEx(pcapPacket);
                }
                if (res == Pcap.NEXT_EX_OK) {
                    received++;

                    PcapPacket pcapPacketCopy = new PcapPacket(pcapPacket); // deep copy
                    pcapPacketCopy.scan(JProtocol.ETHERNET_ID); // parse the headers

                    ingress(pcapPacketCopy);
                }

                // flush the egress queue
                while (!egressPackets.isEmpty()) {
                    JBuffer packet = egressPackets.poll();

                    synchronized (pcapLock) {
                        if (pcap == null) {
                            return;
                        }
                        pcap.sendPacket(packet);
                    }

                    sent++;
                }
            }
        }
    }
}
